@@ -22,6 +22,7 @@ module CmmOpt (
 #include "HsVersions.h"
 
 import Cmm
+import CmmExpr
 import CmmUtils
 import CLabel
 import MachOp
@@ -29,13 +30,12 @@ import StaticFlags
 
 import UniqFM
 import Unique
-
+import FastTypes
 import Outputable
 
 import Data.Bits
 import Data.Word
 import Data.Int
-import GHC.Exts
 
 -- -----------------------------------------------------------------------------
 -- The mini-inliner
@@ -52,6 +52,10 @@ once.  It works as follows:
 	- if we reach the statement that uses it, inline the rhs
 	  and delete the original assignment.
 
+[N.B. In the Quick C-- compiler, this optimization is achieved by a
+ combination of two dataflow passes: forward substitution (peephole
+ optimization) and dead-assignment elimination.  ---NR]
+
 Possible generalisations: here is an example from factorial
 
 Fac_zdwfac_entry:
@@ -85,21 +89,23 @@ To inline _smi:
      its occurrences.
 -}
 
+countUses :: UserOfLocalRegs a => a -> UniqFM Int
+countUses a = foldRegsUsed (\m r -> addToUFM m r (count m r + 1)) emptyUFM a
+  where count m r = lookupWithDefaultUFM m (0::Int) r
+
 cmmMiniInline :: [CmmBasicBlock] -> [CmmBasicBlock]
 cmmMiniInline blocks = map do_inline blocks 
-  where 
-	blockUses (BasicBlock _ stmts)
-	 = foldr (plusUFM_C (+)) emptyUFM (map getStmtUses stmts)
-
-	uses = foldr (plusUFM_C (+)) emptyUFM (map blockUses blocks)
-
-	do_inline (BasicBlock id stmts)
-	 = BasicBlock id (cmmMiniInlineStmts uses stmts)
-
+  where do_inline (BasicBlock id stmts)
+          = BasicBlock id (cmmMiniInlineStmts (countUses blocks) stmts)
 
 cmmMiniInlineStmts :: UniqFM Int -> [CmmStmt] -> [CmmStmt]
 cmmMiniInlineStmts uses [] = []
 cmmMiniInlineStmts uses (stmt@(CmmAssign (CmmLocal (LocalReg u _ _)) expr) : stmts)
+        -- not used at all: just discard this assignment
+  | Nothing <- lookupUFM uses u
+  = cmmMiniInlineStmts uses stmts
+
+        -- used once: try to inline at the use site
   | Just 1 <- lookupUFM uses u,
     Just stmts' <- lookForInline u expr stmts
   = 
@@ -117,7 +123,7 @@ cmmMiniInlineStmts uses (stmt:stmts)
 -- and temporaries are single-assignment.
 lookForInline u expr (stmt@(CmmAssign (CmmLocal (LocalReg u' _ _)) rhs) : rest)
   | u /= u' 
-  = case lookupUFM (getExprUses rhs) u of
+  = case lookupUFM (countUses rhs) u of
 	Just 1 -> Just (inlineStmt u expr stmt : rest)
 	_other -> case lookForInline u expr rest of
 		     Nothing    -> Nothing
@@ -126,8 +132,10 @@ lookForInline u expr (stmt@(CmmAssign (CmmLocal (LocalReg u' _ _)) rhs) : rest)
 lookForInline u expr (CmmNop : rest)
   = lookForInline u expr rest
 
+lookForInline _ _ [] = Nothing
+
 lookForInline u expr (stmt:stmts)
-  = case lookupUFM (getStmtUses stmt) u of
+  = case lookupUFM (countUses stmt) u of
 	Just 1 | ok_to_inline -> Just (inlineStmt u expr stmt : stmts)
 	_other -> Nothing
   where
@@ -140,30 +148,6 @@ lookForInline u expr (stmt:stmts)
 		     CmmCall{} -> hasNoGlobalRegs expr
 		     _ -> True
 
--- -----------------------------------------------------------------------------
--- Boring Cmm traversals for collecting usage info and substitutions.
-
-getStmtUses :: CmmStmt -> UniqFM Int
-getStmtUses (CmmAssign _ e) = getExprUses e
-getStmtUses (CmmStore e1 e2) = plusUFM_C (+) (getExprUses e1) (getExprUses e2)
-getStmtUses (CmmCall target _ es _ _)
-   = plusUFM_C (+) (uses target) (getExprsUses (map fst es))
-   where uses (CmmCallee e _) = getExprUses e
-	 uses _ = emptyUFM
-getStmtUses (CmmCondBranch e _) = getExprUses e
-getStmtUses (CmmSwitch e _) = getExprUses e
-getStmtUses (CmmJump e _) = getExprUses e
-getStmtUses _ = emptyUFM
-
-getExprUses :: CmmExpr -> UniqFM Int
-getExprUses (CmmReg (CmmLocal (LocalReg u _ _))) = unitUFM u 1
-getExprUses (CmmRegOff (CmmLocal (LocalReg u _ _)) _) = unitUFM u 1
-getExprUses (CmmLoad e _) = getExprUses e
-getExprUses (CmmMachOp _ es) = getExprsUses es
-getExprUses _other = emptyUFM
-
-getExprsUses es = foldr (plusUFM_C (+)) emptyUFM (map getExprUses es)
-
 inlineStmt :: Unique -> CmmExpr -> CmmStmt -> CmmStmt
 inlineStmt u a (CmmAssign r e) = CmmAssign r (inlineExpr u a e)
 inlineStmt u a (CmmStore e1 e2) = CmmStore (inlineExpr u a e1) (inlineExpr u a e2)
@@ -171,7 +155,7 @@ inlineStmt u a (CmmCall target regs es srt ret)
    = CmmCall (infn target) regs es' srt ret
    where infn (CmmCallee fn cconv) = CmmCallee fn cconv
 	 infn (CmmPrim p) = CmmPrim p
-	 es' = [ (inlineExpr u a e, hint) | (e,hint) <- es ]
+	 es' = [ (CmmKinded (inlineExpr u a e) hint) | (CmmKinded e hint) <- es ]
 inlineStmt u a (CmmCondBranch e d) = CmmCondBranch (inlineExpr u a e) d
 inlineStmt u a (CmmSwitch e d) = CmmSwitch (inlineExpr u a e) d
 inlineStmt u a (CmmJump e d) = CmmJump (inlineExpr u a e) d
@@ -410,15 +394,15 @@ cmmMachOpFold mop args@[x, y@(CmmLit (CmmInt 0 _))]
     	MO_S_Shr r -> x
     	MO_U_Shr r -> x
         MO_Ne    r | isComparisonExpr x -> x
-	MO_Eq    r | Just x' <- maybeInvertConditionalExpr x -> x'
+	MO_Eq    r | Just x' <- maybeInvertCmmExpr x -> x'
 	MO_U_Gt  r | isComparisonExpr x -> x
 	MO_S_Gt  r | isComparisonExpr x -> x
 	MO_U_Lt  r | isComparisonExpr x -> CmmLit (CmmInt 0 wordRep)
 	MO_S_Lt  r | isComparisonExpr x -> CmmLit (CmmInt 0 wordRep)
 	MO_U_Ge  r | isComparisonExpr x -> CmmLit (CmmInt 1 wordRep)
 	MO_S_Ge  r | isComparisonExpr x -> CmmLit (CmmInt 1 wordRep)
-	MO_U_Le  r | Just x' <- maybeInvertConditionalExpr x -> x'
-	MO_S_Le  r | Just x' <- maybeInvertConditionalExpr x -> x'
+	MO_U_Le  r | Just x' <- maybeInvertCmmExpr x -> x'
+	MO_S_Le  r | Just x' <- maybeInvertCmmExpr x -> x'
     	other    -> CmmMachOp mop args
 
 cmmMachOpFold mop args@[x, y@(CmmLit (CmmInt 1 rep))]
@@ -428,10 +412,10 @@ cmmMachOpFold mop args@[x, y@(CmmLit (CmmInt 1 rep))]
     	MO_U_Quot r -> x
     	MO_S_Rem  r -> CmmLit (CmmInt 0 rep)
     	MO_U_Rem  r -> CmmLit (CmmInt 0 rep)
-        MO_Ne    r | Just x' <- maybeInvertConditionalExpr x -> x'
+        MO_Ne    r | Just x' <- maybeInvertCmmExpr x -> x'
 	MO_Eq    r | isComparisonExpr x -> x
-	MO_U_Lt  r | Just x' <- maybeInvertConditionalExpr x -> x'
-	MO_S_Lt  r | Just x' <- maybeInvertConditionalExpr x -> x'
+	MO_U_Lt  r | Just x' <- maybeInvertCmmExpr x -> x'
+	MO_S_Lt  r | Just x' <- maybeInvertCmmExpr x -> x'
 	MO_U_Gt  r | isComparisonExpr x -> CmmLit (CmmInt 0 wordRep)
 	MO_S_Gt  r | isComparisonExpr x -> CmmLit (CmmInt 0 wordRep)
 	MO_U_Le  r | isComparisonExpr x -> CmmLit (CmmInt 1 wordRep)
@@ -494,42 +478,28 @@ cmmMachOpFold mop args = CmmMachOp mop args
 -- Used to be in MachInstrs --SDM.
 -- ToDo: remove use of unboxery --SDM.
 
-w2i x = word2Int# x
-i2w x = int2Word# x
+-- Unboxery removed in favor of FastInt; but is the function supposed to fail
+-- on inputs >= 2147483648, or was that just an implementation artifact?
+-- And is this speed-critical, or can we just use Integer operations
+-- (including Data.Bits)?
+--  --Isaac Dupree
 
 exactLog2 :: Integer -> Maybe Integer
-exactLog2 x
-  = if (x <= 0 || x >= 2147483648) then
+exactLog2 x_
+  = if (x_ <= 0 || x_ >= 2147483648) then
        Nothing
     else
-       case fromInteger x of { I# x# ->
-       if (w2i ((i2w x#) `and#` (i2w (0# -# x#))) /=# x#) then
+       case iUnbox (fromInteger x_) of { x ->
+       if (x `bitAndFastInt` negateFastInt x) /=# x then
 	  Nothing
        else
-	  Just (toInteger (I# (pow2 x#)))
+	  Just (toInteger (iBox (pow2 x)))
        }
   where
-    pow2 x# | x# ==# 1# = 0#
-            | otherwise = 1# +# pow2 (w2i (i2w x# `shiftRL#` 1#))
+    pow2 x | x ==# _ILIT(1) = _ILIT(0)
+           | otherwise = _ILIT(1) +# pow2 (x `shiftR_FastInt` _ILIT(1))
 
 
--- -----------------------------------------------------------------------------
--- widening / narrowing
-
-narrowU :: MachRep -> Integer -> Integer
-narrowU I8  x = fromIntegral (fromIntegral x :: Word8)
-narrowU I16 x = fromIntegral (fromIntegral x :: Word16)
-narrowU I32 x = fromIntegral (fromIntegral x :: Word32)
-narrowU I64 x = fromIntegral (fromIntegral x :: Word64)
-narrowU _ _ = panic "narrowTo"
-
-narrowS :: MachRep -> Integer -> Integer
-narrowS I8  x = fromIntegral (fromIntegral x :: Int8)
-narrowS I16 x = fromIntegral (fromIntegral x :: Int16)
-narrowS I32 x = fromIntegral (fromIntegral x :: Int32)
-narrowS I64 x = fromIntegral (fromIntegral x :: Int64)
-narrowS _ _ = panic "narrowTo"
-
 -- -----------------------------------------------------------------------------
 -- Loopify for C
 
@@ -584,10 +554,6 @@ isComparisonExpr :: CmmExpr -> Bool
 isComparisonExpr (CmmMachOp op _) = isComparisonMachOp op
 isComparisonExpr _other 	    = False
 
-maybeInvertConditionalExpr :: CmmExpr -> Maybe CmmExpr
-maybeInvertConditionalExpr (CmmMachOp op args) 
-  | Just op' <- maybeInvertComparison op = Just (CmmMachOp op' args)
-maybeInvertConditionalExpr _ = Nothing
-
 isPicReg (CmmReg (CmmGlobal PicBaseReg)) = True
 isPicReg _ = False
+
