@@ -1,4 +1,3 @@
-{-# OPTIONS -fno-warn-missing-signatures #-}
 -----------------------------------------------------------------------------
 --
 -- The register liveness determinator
@@ -6,6 +5,7 @@
 -- (c) The University of Glasgow 2004
 --
 -----------------------------------------------------------------------------
+{-# OPTIONS -Wall -fno-warn-name-shadowing #-}
 
 module RegLiveness (
 	RegSet,
@@ -30,13 +30,12 @@ module RegLiveness (
 
   ) where
 
-#include "HsVersions.h"
-
+import BlockId
 import MachRegs
 import MachInstrs
 import PprMach
 import RegAllocInfo
-import Cmm
+import Cmm hiding (RegSet)
 
 import Digraph
 import Outputable
@@ -46,6 +45,7 @@ import UniqFM
 import UniqSupply
 import Bag
 import State
+import FastString
 
 import Data.List
 import Data.Maybe
@@ -112,9 +112,9 @@ instance Outputable LiveInstr where
 	 =  ppr instr
 		$$ (nest 8
 			$ vcat
-			[ pprRegs (ptext SLIT("# born:    ")) (liveBorn live)
-			, pprRegs (ptext SLIT("# r_dying: ")) (liveDieRead live)
-			, pprRegs (ptext SLIT("# w_dying: ")) (liveDieWrite live) ]
+			[ pprRegs (ptext (sLit "# born:    ")) (liveBorn live)
+			, pprRegs (ptext (sLit "# r_dying: ")) (liveDieRead live)
+			, pprRegs (ptext (sLit "# w_dying: ")) (liveDieWrite live) ]
 		    $+$ space)
 
 	 where 	pprRegs :: SDoc -> RegSet -> SDoc
@@ -154,6 +154,7 @@ mapBlockTopM f (CmmProc header label params (ListGraph comps))
  = do	comps'	<- mapM (mapBlockCompM f) comps
  	return	$ CmmProc header label params (ListGraph comps')
 
+mapBlockCompM :: Monad m => (a -> m a') -> (GenBasicBlock a) -> m (GenBasicBlock a')
 mapBlockCompM f (BasicBlock i blocks)
  = do	blocks'	<- mapM f blocks
  	return	$ BasicBlock i blocks'
@@ -204,7 +205,7 @@ slurpConflicts live
 		= (consBag rsLiveEntry conflicts, moves)
 
 		| otherwise
-		= error "RegLiveness.slurpBlock: bad block"
+		= panic "RegLiveness.slurpBlock: bad block"
 
 	slurpLIs rsLive (conflicts, moves) []
 		= (consBag rsLive conflicts, moves)
@@ -243,14 +244,14 @@ slurpConflicts live
 					, moves) lis
 
 
--- | For spill/reloads
+-- | For spill\/reloads
 --
 --	SPILL  v1, slot1
 --	...
 --	RELOAD slot1, v2
 --
 --	If we can arrange that v1 and v2 are allocated to the same hreg it's more likely
---	the spill/reload instrs can be cleaned and replaced by a nop reg-reg move.
+--	the spill\/reload instrs can be cleaned and replaced by a nop reg-reg move.
 --
 --
 slurpReloadCoalesce :: LiveCmmTop -> Bag (Reg, Reg)
@@ -501,7 +502,7 @@ regLiveness (CmmProc info lbl params (ListGraph blocks@(first : _)))
 
 
 sccBlocks :: [NatBasicBlock] -> [SCC NatBasicBlock]
-sccBlocks blocks = stronglyConnComp graph
+sccBlocks blocks = stronglyConnCompFromEdgedVertices graph
   where
 	getOutEdges :: [Instr] -> [BlockId]
 	getOutEdges instrs = foldl' (\a x -> jumpDests x a) [] instrs
@@ -557,7 +558,7 @@ livenessSCCs blockmap done
 	          concatMap tail $
 	          groupBy (\(a1, _) (a2, _) -> eq a1 a2) $
 	          iterate (\(a, _) -> f a b) $
-	          (a, error "RegisterAlloc.livenessSCCs")
+	          (a, panic "RegLiveness.livenessSCCs")
 
 
             linearLiveness :: BlockMap RegSet -> [NatBasicBlock]
@@ -640,6 +641,7 @@ livenessBack liveregs blockmap acc (instr : instrs)
    in	livenessBack liveregs' blockmap (instr' : acc) instrs
 
 -- don't bother tagging comments or deltas with liveness
+liveness1 :: RegSet -> BlockMap RegSet -> Instr -> (RegSet, LiveInstr)
 liveness1 liveregs _   (instr@COMMENT{})
 	= (liveregs, Instr instr Nothing)
 
