@@ -1,16 +1,10 @@
-{-# OPTIONS -w #-}
--- The above warning supression flag is a temporary kludge.
--- While working on this module you are encouraged to remove it and fix
--- any warnings in the module. See
---     http://hackage.haskell.org/trac/ghc/wiki/Commentary/CodingStyle#Warnings
--- for details
 
-module VectType ( vectTyCon, vectType, vectTypeEnv,
-                   PAInstance, buildPADict )
+module VectType ( vectTyCon, vectAndLiftType, vectType, vectTypeEnv,
+                  mkRepr, arrShapeTys, arrShapeVars, arrSelector,
+                  buildPADict,
+                  fromVect )
 where
 
-#include "HsVersions.h"
-
 import VectMonad
 import VectUtils
 import VectCore
@@ -25,11 +19,10 @@ import Type
 import TypeRep
 import Coercion
 import FamInstEnv        ( FamInst, mkLocalFamInst )
-import InstEnv           ( Instance, mkLocalInstance, instanceDFunId )
 import OccName
 import MkId
-import BasicTypes        ( StrictnessMark(..), OverlapFlag(..), boolToRecFlag )
-import Var               ( Var )
+import BasicTypes        ( StrictnessMark(..), boolToRecFlag )
+import Var               ( Var, TyVar )
 import Id                ( mkWildId )
 import Name              ( Name, getOccName )
 import NameEnv
@@ -40,9 +33,10 @@ import Unique
 import UniqFM
 import UniqSet
 import Util
-import Digraph           ( SCC(..), stronglyConnComp )
+import Digraph           ( SCC(..), stronglyConnCompFromEdgedVertices )
 
 import Outputable
+import FastString
 
 import Control.Monad  ( liftM, liftM2, zipWithM, zipWithM_, mapAndUnzipM )
 import Data.List      ( inits, tails, zipWith4, zipWith5 )
@@ -55,13 +49,22 @@ vectTyCon tc
   | isFunTyCon tc        = builtin closureTyCon
   | isBoxedTupleTyCon tc = return tc
   | isUnLiftedTyCon tc   = return tc
-  | otherwise = do
-                  r <- lookupTyCon tc
-                  case r of
-                    Just tc' -> return tc'
+  | otherwise            = maybeCantVectoriseM "Tycon not vectorised:" (ppr tc)
+                         $ lookupTyCon tc
+
+vectAndLiftType :: Type -> VM (Type, Type)
+vectAndLiftType ty | Just ty' <- coreView ty = vectAndLiftType ty'
+vectAndLiftType ty
+  = do
+      mdicts   <- mapM paDictArgType tyvars
+      let dicts = [dict | Just dict <- mdicts]
+      vmono_ty <- vectType mono_ty
+      lmono_ty <- mkPArrayType vmono_ty
+      return (abstractType tyvars dicts vmono_ty,
+              abstractType tyvars dicts lmono_ty)
+  where
+    (tyvars, mono_ty) = splitForAllTys ty
 
-                    -- FIXME: just for now
-                    Nothing  -> pprTrace "ccTyCon:" (ppr tc) $ return tc
 
 vectType :: Type -> VM Type
 vectType ty | Just ty' <- coreView ty = vectType ty'
@@ -69,29 +72,42 @@ vectType (TyVarTy tv) = return $ TyVarTy tv
 vectType (AppTy ty1 ty2) = liftM2 AppTy (vectType ty1) (vectType ty2)
 vectType (TyConApp tc tys) = liftM2 TyConApp (vectTyCon tc) (mapM vectType tys)
 vectType (FunTy ty1 ty2)   = liftM2 TyConApp (builtin closureTyCon)
-                                             (mapM vectType [ty1,ty2])
+                                             (mapM vectAndBoxType [ty1,ty2])
 vectType ty@(ForAllTy _ _)
   = do
       mdicts   <- mapM paDictArgType tyvars
       mono_ty' <- vectType mono_ty
-      return $ tyvars `mkForAllTys` ([dict | Just dict <- mdicts] `mkFunTys` mono_ty')
+      return $ abstractType tyvars [dict | Just dict <- mdicts] mono_ty'
   where
     (tyvars, mono_ty) = splitForAllTys ty
 
-vectType ty = pprPanic "vectType:" (ppr ty)
+vectType ty = cantVectorise "Can't vectorise type" (ppr ty)
+
+vectAndBoxType :: Type -> VM Type
+vectAndBoxType ty = vectType ty >>= boxType
+
+abstractType :: [TyVar] -> [Type] -> Type -> Type
+abstractType tyvars dicts = mkForAllTys tyvars . mkFunTys dicts
+
+-- ----------------------------------------------------------------------------
+-- Boxing
+
+boxType :: Type -> VM Type
+boxType ty
+  | Just (tycon, []) <- splitTyConApp_maybe ty
+  , isUnLiftedTyCon tycon
+  = do
+      r <- lookupBoxedTyCon tycon
+      case r of
+        Just tycon' -> return $ mkTyConApp tycon' []
+        Nothing     -> return ty
+boxType ty = return ty
 
 -- ----------------------------------------------------------------------------
 -- Type definitions
 
 type TyConGroup = ([TyCon], UniqSet TyCon)
 
-data PAInstance = PAInstance {
-                    painstDFun      :: Var
-                  , painstOrigTyCon :: TyCon
-                  , painstVectTyCon :: TyCon
-                  , painstArrTyCon  :: TyCon
-                  }
-
 vectTypeEnv :: TypeEnv -> VM (TypeEnv, [FamInst], [(Var, CoreExpr)])
 vectTypeEnv env
   = do
@@ -129,10 +145,6 @@ vectTypeEnv env
 
     mk_map env = listToUFM_Directly [(u, getUnique n /= u) | (u,n) <- nameEnvUniqueElts env]
 
-    keep_tc tc = let dcs = tyConDataCons tc
-                 in
-                 defTyCon tc tc >> zipWithM_ defDataCon dcs dcs
-
 
 vectTyConDecls :: [TyCon] -> VM [TyCon]
 vectTyConDecls tcs = fixV $ \tcs' ->
@@ -144,7 +156,7 @@ vectTyConDecl :: TyCon -> VM TyCon
 vectTyConDecl tc
   = do
       name' <- cloneName mkVectTyConOcc name
-      rhs'  <- vectAlgTyConRhs (algTyConRhs tc)
+      rhs'  <- vectAlgTyConRhs tc (algTyConRhs tc)
 
       liftDs $ buildAlgTyCon name'
                              tyvars
@@ -159,21 +171,24 @@ vectTyConDecl tc
     tyvars = tyConTyVars tc
     rec_flag = boolToRecFlag (isRecursiveTyCon tc)
 
-vectAlgTyConRhs :: AlgTyConRhs -> VM AlgTyConRhs
-vectAlgTyConRhs (DataTyCon { data_cons = data_cons
-                           , is_enum   = is_enum
-                           })
+vectAlgTyConRhs :: TyCon -> AlgTyConRhs -> VM AlgTyConRhs
+vectAlgTyConRhs _ (DataTyCon { data_cons = data_cons
+                             , is_enum   = is_enum
+                             })
   = do
       data_cons' <- mapM vectDataCon data_cons
       zipWithM_ defDataCon data_cons data_cons'
       return $ DataTyCon { data_cons = data_cons'
                          , is_enum   = is_enum
                          }
+vectAlgTyConRhs tc _ = cantVectorise "Can't vectorise type definition:" (ppr tc)
 
 vectDataCon :: DataCon -> VM DataCon
 vectDataCon dc
-  | not . null $ dataConExTyVars dc = pprPanic "vectDataCon: existentials" (ppr dc)
-  | not . null $ dataConEqSpec   dc = pprPanic "vectDataCon: eq spec" (ppr dc)
+  | not . null $ dataConExTyVars dc
+        = cantVectorise "Can't vectorise constructor (existentials):" (ppr dc)
+  | not . null $ dataConEqSpec   dc
+        = cantVectorise "Can't vectorise constructor (eq spec):" (ppr dc)
   | otherwise
   = do
       name'    <- cloneName mkVectDataConOcc name
@@ -209,6 +224,7 @@ buildPReprTyCon orig_tc vect_tc
       liftDs $ buildSynTyCon name
                              tyvars
                              (SynonymTyCon rhs_ty)
+			     (typeKind rhs_ty)
                              (Just $ mk_fam_inst prepr_tc vect_tc)
   where
     tyvars = tyConTyVars vect_tc
@@ -253,10 +269,13 @@ voidRepr
                , void_bottom = Var var
                }
 
+{-
 enumRepr :: VM Repr
 enumRepr
   = do
-      (arr_tycon, _) <- parrayReprTyCon intTy
+      tycon <- builtin enumerationTyCon
+      let [data_con] = tyConDataCons tycon
+      (arr_tycon, _) <- parrayReprTyCon (mkTyConApp tycon [])
       let [arr_data_con] = tyConDataCons arr_tycon
 
       return $ EnumRepr {
@@ -265,9 +284,7 @@ enumRepr
                , enum_arr_tycon    = arr_tycon
                , enum_arr_data_con = arr_data_con
                }
-  where
-    tycon = intTyCon
-    data_con = intDataCon
+-}
 
 unboxedProductRepr :: [Type] -> VM Repr
 unboxedProductRepr []   = voidRepr
@@ -280,7 +297,8 @@ boxedProductRepr tys
       tycon <- builtin (prodTyCon arity)
       let [data_con] = tyConDataCons tycon
 
-      (arr_tycon, _) <- parrayReprTyCon $ mkTyConApp tycon tys
+      tys' <- mapM boxType tys
+      (arr_tycon, _) <- parrayReprTyCon $ mkTyConApp tycon tys'
       let [arr_data_con] = tyConDataCons arr_tycon
 
       return $ ProdRepr {
@@ -336,28 +354,39 @@ arrReprType :: Repr -> VM Type
 arrReprType = mkPArrayType . reprType
 
 arrShapeTys :: Repr -> VM [Type]
-arrShapeTys (SumRepr  {})
-  = do
-      int_arr <- builtin parrayIntPrimTyCon
-      return [intPrimTy, mkTyConApp int_arr [], mkTyConApp int_arr []]
+arrShapeTys (SumRepr  {}) = sumShapeTys
 arrShapeTys (ProdRepr {}) = return [intPrimTy]
 arrShapeTys (IdRepr _)    = return []
 arrShapeTys (VoidRepr {}) = return [intPrimTy]
-arrShapeTys (EnumRepr {}) = return [intPrimTy]
+arrShapeTys (EnumRepr {}) = sumShapeTys
+
+sumShapeTys :: VM [Type]
+sumShapeTys = do
+                int_arr <- builtin intPrimArrayTy
+                return [intPrimTy, int_arr, int_arr]
+
 
 arrShapeVars :: Repr -> VM [Var]
-arrShapeVars repr = mapM (newLocalVar FSLIT("sh")) =<< arrShapeTys repr
+arrShapeVars repr = mapM (newLocalVar (fsLit "sh")) =<< arrShapeTys repr
 
 replicateShape :: Repr -> CoreExpr -> CoreExpr -> VM [CoreExpr]
-replicateShape (ProdRepr {}) len _ = return [len]
-replicateShape (SumRepr {})  len tag
+replicateShape (ProdRepr {}) len _   = return [len]
+replicateShape (SumRepr {})  len tag = replicateSumShape len tag
+replicateShape (IdRepr _) _ _        = return []
+replicateShape (VoidRepr {}) len _   = return [len]
+replicateShape (EnumRepr {}) len tag = replicateSumShape len tag
+
+replicateSumShape :: CoreExpr -> CoreExpr -> VM [CoreExpr]
+replicateSumShape len tag
   = do
       rep <- builtin replicatePAIntPrimVar
       up  <- builtin upToPAIntPrimVar
       return [len, Var rep `mkApps` [len, tag], Var up `App` len]
-replicateShape (IdRepr _) _ _ = return []
-replicateShape (VoidRepr {}) len _ = return [len]
-replicateShape (EnumRepr {}) len _ = return [len]
+
+arrSelector :: Repr -> [CoreExpr] -> VM (CoreExpr, CoreExpr, CoreExpr)
+arrSelector (SumRepr {})  [len, sel, is] = return (len, sel, is)
+arrSelector (EnumRepr {}) [len, sel, is] = return (len, sel, is)
+arrSelector _             _              = panic "arrSelector"
 
 emptyArrRepr :: Repr -> VM [CoreExpr]
 emptyArrRepr (SumRepr { sum_components = prods })
@@ -370,8 +399,8 @@ emptyArrRepr (IdRepr ty)
   = liftM singleton $ emptyPA ty
 emptyArrRepr (VoidRepr { void_tycon = tycon })
   = liftM singleton $ emptyPA (mkTyConApp tycon [])
-emptyArrRepr (EnumRepr { enum_tycon = tycon })
-  = liftM singleton $ emptyPA (mkTyConApp tycon [])
+emptyArrRepr (EnumRepr {})
+  = return []
 
 arrReprTys :: Repr -> VM [Type]
 arrReprTys (SumRepr { sum_components = reprs })
@@ -385,7 +414,7 @@ arrReprTys (IdRepr ty)
 arrReprTys (VoidRepr { void_tycon = tycon })
   = liftM singleton $ mkPArrayType (mkTyConApp tycon [])
 arrReprTys (EnumRepr {})
-  = liftM singleton $ mkPArrayType intPrimTy
+  = return []
 
 arrReprTys' :: Repr -> VM [[Type]]
 arrReprTys' (SumRepr { sum_components = reprs })
@@ -394,12 +423,12 @@ arrReprTys' repr = liftM singleton $ arrReprTys repr
 
 arrReprVars :: Repr -> VM [[Var]]
 arrReprVars repr
-  = mapM (mapM (newLocalVar FSLIT("rs"))) =<< arrReprTys' repr
+  = mapM (mapM (newLocalVar (fsLit "rs"))) =<< arrReprTys' repr
 
 mkRepr :: TyCon -> VM Repr
 mkRepr vect_tc
   | [tys] <- rep_tys = boxedProductRepr tys
-  | all null rep_tys = enumRepr
+  -- removed: | all null rep_tys = enumRepr
   | otherwise        = sumRepr =<< mapM unboxedProductRepr rep_tys
   where
     rep_tys = map dataConRepArgTys $ tyConDataCons vect_tc
@@ -410,7 +439,7 @@ buildPReprType = liftM reprType . mkRepr
 buildToPRepr :: Repr -> TyCon -> TyCon -> TyCon -> VM CoreExpr
 buildToPRepr repr vect_tc prepr_tc _
   = do
-      arg    <- newLocalVar FSLIT("x") arg_ty
+      arg    <- newLocalVar (fsLit "x") arg_ty
       result <- to_repr repr (Var arg)
 
       return . Lam arg
@@ -452,23 +481,25 @@ buildToPRepr repr vect_tc prepr_tc _
     to_unboxed (ProdRepr { prod_components = tys
                          , prod_data_con   = data_con })
       = do
-          vars <- mapM (newLocalVar FSLIT("r")) tys
+          vars <- mapM (newLocalVar (fsLit "r")) tys
           return (vars, mkConApp data_con (map Type tys ++ map Var vars))
 
     to_unboxed (IdRepr ty)
       = do
-          var <- newLocalVar FSLIT("y") ty
+          var <- newLocalVar (fsLit "y") ty
           return ([var], Var var)
 
     to_unboxed (VoidRepr { void_bottom = bottom })
       = return ([], bottom)
 
+    to_unboxed _ = panic "buildToPRepr/to_unboxed"
+
 
 buildFromPRepr :: Repr -> TyCon -> TyCon -> TyCon -> VM CoreExpr
 buildFromPRepr repr vect_tc prepr_tc _
   = do
       arg_ty <- mkPReprType res_ty
-      arg    <- newLocalVar FSLIT("x") arg_ty
+      arg    <- newLocalVar (fsLit "x") arg_ty
 
       liftM (Lam arg)
            . from_repr repr
@@ -484,7 +515,7 @@ buildFromPRepr repr vect_tc prepr_tc _
                             , sum_tycon      = tycon })
               expr
       = do
-          vars   <- mapM (newLocalVar FSLIT("x")) (map reprType prods)
+          vars   <- mapM (newLocalVar (fsLit "x")) (map reprType prods)
           bodies <- sequence . zipWith3 from_unboxed prods cons
                              $ map Var vars
           return . Case expr (mkWildId (reprType repr)) res_ty
@@ -494,7 +525,7 @@ buildFromPRepr repr vect_tc prepr_tc _
 
     from_repr repr@(EnumRepr { enum_data_con = data_con }) expr
       = do
-          var <- newLocalVar FSLIT("n") intPrimTy
+          var <- newLocalVar (fsLit "n") intPrimTy
 
           let res = Case (Var var) (mkWildId intPrimTy) res_ty
                   $ (DEFAULT, [], error_expr)
@@ -516,21 +547,23 @@ buildFromPRepr repr vect_tc prepr_tc _
               con
               expr
       = do
-          vars <- mapM (newLocalVar FSLIT("y")) tys
+          vars <- mapM (newLocalVar (fsLit "y")) tys
           return $ Case expr (mkWildId (reprType prod)) res_ty
                    [(DataAlt data_con, vars, con `mkVarApps` vars)]
 
     from_unboxed (IdRepr _) con expr
        = return $ con `App` expr
 
-    from_unboxed (VoidRepr {}) con expr
+    from_unboxed (VoidRepr {}) con _
        = return con
 
+    from_unboxed _ _ _ = panic "buildFromPRepr/from_unboxed"
+
 buildToArrPRepr :: Repr -> TyCon -> TyCon -> TyCon -> VM CoreExpr
 buildToArrPRepr repr vect_tc prepr_tc arr_tc
   = do
       arg_ty     <- mkPArrayType el_ty
-      arg        <- newLocalVar FSLIT("xs") arg_ty
+      arg        <- newLocalVar (fsLit "xs") arg_ty
 
       res_ty     <- mkPArrayType (reprType repr)
 
@@ -558,7 +591,7 @@ buildToArrPRepr repr vect_tc prepr_tc arr_tc
 
     [arr_dc] = tyConDataCons arr_tc
 
-    to_repr shape_vars@(len_var : _)
+    to_repr shape_vars@(_ : _)
             repr_vars
             (SumRepr { sum_components   = prods
                      , sum_arr_tycon    = tycon
@@ -581,12 +614,15 @@ buildToArrPRepr repr vect_tc prepr_tc arr_tc
                 . mkConApp data_con
                 $ map Type tys ++ map Var (len_var : repr_vars)
 
-    to_repr [len_var]
-            [[repr_var]]
+    to_repr shape_vars
+            _
             (EnumRepr { enum_arr_tycon    = tycon
                       , enum_arr_data_con = data_con })
       = return . wrapFamInstBody tycon []
-               $ mkConApp data_con [Var len_var, Var repr_var]
+               . mkConApp data_con
+                $ map Var shape_vars
+
+    to_repr _ _ _ = panic "buildToArrPRepr/to_repr"
 
     to_prod repr_vars@(r : _)
             (ProdRepr { prod_components   = tys@(ty : _)
@@ -598,15 +634,16 @@ buildToArrPRepr repr vect_tc prepr_tc arr_tc
                  . mkConApp data_con
                  $ map Type tys ++ len : map Var repr_vars
 
-    to_prod [var] (IdRepr ty)   = return (Var var)
+    to_prod [var] (IdRepr _)    = return (Var var)
     to_prod [var] (VoidRepr {}) = return (Var var)
+    to_prod _     _             = panic "buildToArrPRepr/to_prod"
 
 
 buildFromArrPRepr :: Repr -> TyCon -> TyCon -> TyCon -> VM CoreExpr
 buildFromArrPRepr repr vect_tc prepr_tc arr_tc
   = do
       arg_ty     <- mkPArrayType =<< mkPReprType el_ty
-      arg        <- newLocalVar FSLIT("xs") arg_ty
+      arg        <- newLocalVar (fsLit "xs") arg_ty
 
       res_ty     <- mkPArrayType el_ty
 
@@ -642,7 +679,7 @@ buildFromArrPRepr repr vect_tc prepr_tc arr_tc
               res_ty
               body
       = do
-          vars <- mapM (newLocalVar FSLIT("xs")) =<< mapM arrReprType prods
+          vars <- mapM (newLocalVar (fsLit "xs")) =<< mapM arrReprType prods
           result <- go prods repr_vars vars body
 
           let scrut = unwrapFamInstScrut tycon ty_args expr
@@ -655,17 +692,20 @@ buildFromArrPRepr repr vect_tc prepr_tc arr_tc
         go [] [] [] body = return body
         go (prod : prods) (repr_vars : rss) (var : vars) body
           = do
-              shape_vars <- mapM (newLocalVar FSLIT("s")) =<< arrShapeTys prod
+              shape_vars <- mapM (newLocalVar (fsLit "s")) =<< arrShapeTys prod
 
               from_prod prod (Var var) shape_vars repr_vars res_ty
                 =<< go prods rss vars body
+        go _ _ _ _ = panic "buildFromArrPRepr/go"
 
     from_repr repr expr shape_vars [repr_vars] res_ty body
       = from_prod repr expr shape_vars repr_vars res_ty body
 
-    from_prod prod@(ProdRepr { prod_components = tys
-                             , prod_arr_tycon  = tycon
-                             , prod_arr_data_con = data_con })
+    from_repr _ _ _ _ _ _ = panic "buildFromArrPRepr/from_repr"
+
+    from_prod (ProdRepr { prod_components = tys
+                        , prod_arr_tycon  = tycon
+                        , prod_arr_data_con = data_con })
               expr
               shape_vars
               repr_vars
@@ -674,7 +714,6 @@ buildFromArrPRepr repr vect_tc prepr_tc arr_tc
       = do
           let scrut    = unwrapFamInstScrut tycon tys expr
               scrut_ty = mkTyConApp tycon tys
-          ty <- arrReprType prod
 
           return $ Case scrut (mkWildId scrut_ty) res_ty
                    [(DataAlt data_con, shape_vars ++ repr_vars, body)]
@@ -682,32 +721,34 @@ buildFromArrPRepr repr vect_tc prepr_tc arr_tc
     from_prod (EnumRepr { enum_arr_tycon = tycon
                         , enum_arr_data_con = data_con })
               expr
-              [len_var]
-              [repr_var]
+              shape_vars
+              _
               res_ty
               body
       = let scrut    = unwrapFamInstScrut tycon [] expr
             scrut_ty = mkTyConApp tycon []
         in
         return $ Case scrut (mkWildId scrut_ty) res_ty
-                 [(DataAlt data_con, [len_var, repr_var], body)]
+                 [(DataAlt data_con, shape_vars, body)]
 
-    from_prod (IdRepr ty)
+    from_prod (IdRepr _)
               expr
-              shape_vars
+              _shape_vars
               [repr_var]
-              res_ty
+              _res_ty
               body
       = return $ Let (NonRec repr_var expr) body
 
     from_prod (VoidRepr {})
               expr
-              shape_vars
+              _shape_vars
               [repr_var]
-              res_ty
+              _res_ty
               body
       = return $ Let (NonRec repr_var expr) body
 
+    from_prod _ _ _ _ _ _ = panic "buildFromArrPRepr/from_prod"
+
 buildPRDictRepr :: Repr -> VM CoreExpr
 buildPRDictRepr (VoidRepr { void_tycon = tycon })
   = prDFunOfTyCon tycon
@@ -810,12 +851,6 @@ buildTyConBindings orig_tc vect_tc prepr_tc arr_tc dfun
       dict <- buildPADict repr vect_tc prepr_tc arr_tc dfun
       binds <- takeHoisted
       return $ (dfun, dict) : binds
-  where
-    orig_dcs = tyConDataCons orig_tc
-    vect_dcs = tyConDataCons vect_tc
-    [arr_dc] = tyConDataCons arr_tc
-
-    repr_tys = map dataConRepArgTys vect_dcs
 
 vectDataConWorkers :: Repr -> TyCon -> TyCon -> TyCon
                    -> VM ()
@@ -848,7 +883,7 @@ vectDataConWorkers repr orig_tc vect_tc arr_tc
     lift_data_con tys pre_reprs post_reprs tag
       = do
           len  <- builtin liftingContext
-          args <- mapM (newLocalVar FSLIT("xs"))
+          args <- mapM (newLocalVar (fsLit "xs"))
                   =<< mapM mkPArrayType tys
 
           shape <- replicateShape repr (Var len) tag
@@ -867,7 +902,7 @@ vectDataConWorkers repr orig_tc vect_tc arr_tc
           units <- replicatePA len (Var unitDataConId)
           return [units]
 
-    mk_arr_repr len arrs = return arrs
+    mk_arr_repr _ arrs = return arrs
 
     def_worker data_con arg_tys mk_body
       = do
@@ -884,7 +919,7 @@ vectDataConWorkers repr orig_tc vect_tc arr_tc
         orig_worker = dataConWorkId data_con
 
 buildPADict :: Repr -> TyCon -> TyCon -> TyCon -> Var -> VM CoreExpr
-buildPADict repr vect_tc prepr_tc arr_tc dfun
+buildPADict repr vect_tc prepr_tc arr_tc _
   = polyAbstract tvs $ \abstract ->
     do
       meth_binds <- mapM (mk_method repr) paMethods
@@ -905,11 +940,12 @@ buildPADict repr vect_tc prepr_tc arr_tc dfun
           var  <- newLocalVar name (exprType body)
           return (var, mkInlineMe body)
 
-paMethods = [(FSLIT("toPRepr"),      buildToPRepr),
-             (FSLIT("fromPRepr"),    buildFromPRepr),
-             (FSLIT("toArrPRepr"),   buildToArrPRepr),
-             (FSLIT("fromArrPRepr"), buildFromArrPRepr),
-             (FSLIT("dictPRepr"),    buildPRDict)]
+paMethods :: [(FastString, Repr -> TyCon -> TyCon -> TyCon -> VM CoreExpr)]
+paMethods = [(fsLit "toPRepr",      buildToPRepr),
+             (fsLit "fromPRepr",    buildFromPRepr),
+             (fsLit "toArrPRepr",   buildToArrPRepr),
+             (fsLit "fromArrPRepr", buildFromArrPRepr),
+             (fsLit "dictPRepr",    buildPRDict)]
 
 -- | Split the given tycons into two sets depending on whether they have to be
 -- converted (first list) or not (second list). The first argument contains
@@ -922,7 +958,7 @@ paMethods = [(FSLIT("toPRepr"),      buildToPRepr),
 classifyTyCons :: UniqFM Bool -> [TyConGroup] -> ([TyCon], [TyCon])
 classifyTyCons = classify [] []
   where
-    classify conv keep cs [] = (conv, keep)
+    classify conv keep _  [] = (conv, keep)
     classify conv keep cs ((tcs, ds) : rs)
       | can_convert && must_convert
         = classify (tcs ++ conv) keep (cs `addListToUFM` [(tc,True) | tc <- tcs]) rs
@@ -941,7 +977,7 @@ classifyTyCons = classify [] []
 -- | Compute mutually recursive groups of tycons in topological order
 --
 tyConGroups :: [TyCon] -> [TyConGroup]
-tyConGroups tcs = map mk_grp (stronglyConnComp edges)
+tyConGroups tcs = map mk_grp (stronglyConnCompFromEdgedVertices edges)
   where
     edges = [((tc, ds), tc, uniqSetToList ds) | tc <- tcs
                                 , let ds = tyConsOfTyCon tc]
@@ -958,7 +994,7 @@ tyConsOfTyCon
 tyConsOfType :: Type -> UniqSet TyCon
 tyConsOfType ty
   | Just ty' <- coreView ty    = tyConsOfType ty'
-tyConsOfType (TyVarTy v)       = emptyUniqSet
+tyConsOfType (TyVarTy _)       = emptyUniqSet
 tyConsOfType (TyConApp tc tys) = extend (tyConsOfTypes tys)
   where
     extend | isUnLiftedTyCon tc
@@ -975,3 +1011,41 @@ tyConsOfType other             = pprPanic "ClosureConv.tyConsOfType" $ ppr other
 tyConsOfTypes :: [Type] -> UniqSet TyCon
 tyConsOfTypes = unionManyUniqSets . map tyConsOfType
 
+
+-- ----------------------------------------------------------------------------
+-- Conversions
+
+fromVect :: Type -> CoreExpr -> VM CoreExpr
+fromVect ty expr | Just ty' <- coreView ty = fromVect ty' expr
+fromVect (FunTy arg_ty res_ty) expr
+  = do
+      arg     <- newLocalVar (fsLit "x") arg_ty
+      varg    <- toVect arg_ty (Var arg)
+      varg_ty <- vectType arg_ty
+      vres_ty <- vectType res_ty
+      apply   <- builtin applyClosureVar
+      body    <- fromVect res_ty
+               $ Var apply `mkTyApps` [varg_ty, vres_ty] `mkApps` [expr, varg]
+      return $ Lam arg body
+fromVect ty expr
+  = identityConv ty >> return expr
+
+toVect :: Type -> CoreExpr -> VM CoreExpr
+toVect ty expr = identityConv ty >> return expr
+
+identityConv :: Type -> VM ()
+identityConv ty | Just ty' <- coreView ty = identityConv ty'
+identityConv (TyConApp tycon tys)
+  = do
+      mapM_ identityConv tys
+      identityConvTyCon tycon
+identityConv _ = noV
+
+identityConvTyCon :: TyCon -> VM ()
+identityConvTyCon tc
+  | isBoxedTupleTyCon tc = return ()
+  | isUnLiftedTyCon tc   = return ()
+  | otherwise            = do
+                             tc' <- maybeV (lookupTyCon tc)
+                             if tc == tc' then return () else noV
+
