@@ -1,4 +1,11 @@
-{-# OPTIONS_GHC -fno-bang-patterns #-}
+{-# OPTIONS_GHC -XNoBangPatterns -fno-warn-unused-imports #-}
+{-# OPTIONS_HADDOCK hide #-}
+-- XXX With a GHC 6.9 we get a spurious
+-- Data/Array/Base.hs:26:0:
+--     Warning: Module `Data.Ix' is imported, but nothing from it is used,
+--                except perhaps instances visible in `Data.Ix'
+--              To suppress this warning, use: import Data.Ix()
+-- The -fno-warn-unused-imports works around that bug
 
 -----------------------------------------------------------------------------
 -- |
@@ -18,15 +25,10 @@
 -- #hide
 module Data.Array.Base where
 
-import Prelude
-
 import Control.Monad.ST.Lazy ( strictToLazyST )
 import qualified Control.Monad.ST.Lazy as Lazy (ST)
-import Data.Ix		( Ix, range, index, rangeSize )
-import Data.Int
-import Data.Word
+import Data.Ix ( Ix, range, index, rangeSize )
 import Foreign.C.Types
-import Foreign.Ptr
 import Foreign.StablePtr
 
 #ifdef __GLASGOW_HASKELL__
@@ -41,7 +43,12 @@ import GHC.Float	( Float(..), Double(..) )
 import GHC.Stable	( StablePtr(..) )
 import GHC.Int		( Int8(..),  Int16(..),  Int32(..),  Int64(..) )
 import GHC.Word		( Word8(..), Word16(..), Word32(..), Word64(..) )
-import GHC.IOBase       ( IO(..) )
+import GHC.IOBase       ( IO(..), IOArray(..), stToIO,
+                          newIOArray, unsafeReadIOArray, unsafeWriteIOArray )
+#else
+import Data.Int
+import Data.Word
+import Foreign.Ptr
 #endif
 
 #ifdef __HUGS__
@@ -50,6 +57,7 @@ import Foreign.Storable
 import qualified Hugs.Array as Arr
 import qualified Hugs.ST as ArrST
 import Hugs.Array ( unsafeIndex )
+import Hugs.IOArray
 import Hugs.ST ( STArray, ST(..), runST )
 import Hugs.ByteArray
 #endif
@@ -143,7 +151,7 @@ in the list is out of bounds.  If any two associations in the list have
 the same index, the value at that index is implementation-dependent.
 (In GHC, the last value specified for that index is used.
 Other implementations will also do this for unboxed arrays, but Haskell
-98 requires that for 'Data.Array.Array' the value at such indices is bottom.)
+98 requires that for 'Array' the value at such indices is bottom.)
 
 Because the indices must be checked for these errors, 'array' is
 strict in the bounds argument and in the indices of the association
@@ -295,7 +303,7 @@ indices arr = case bounds arr of (l,u) -> range (l,u)
 -- as their indices.
 elems :: (IArray a e, Ix i) => a i e -> [e]
 elems arr = case bounds arr of
-    (l,u) -> [unsafeAt arr i | i <- [0 .. numElements arr - 1]]
+    (_l, _u) -> [unsafeAt arr i | i <- [0 .. numElements arr - 1]]
 
 {-# INLINE assocs #-}
 -- | Returns the contents of an array as a list of associations.
@@ -324,9 +332,9 @@ accumArray :: (IArray a e, Ix i)
            -> (i,i)              -- ^ The bounds of the array
            -> [(i, e')]          -- ^ List of associations
            -> a i e              -- ^ Returns: the array
-accumArray f init (l,u) ies =
+accumArray f initialValue (l,u) ies =
     let n = safeRangeSize (l, u)
-    in unsafeAccumArray f init (l,u)
+    in unsafeAccumArray f initialValue (l,u)
                         [(safeIndex (l,u) n i, e) | (i, e) <- ies]
 
 {-# INLINE (//) #-}
@@ -341,7 +349,7 @@ As with the 'array' function, if any two associations in the list have
 the same index, the value at that index is implementation-dependent.
 (In GHC, the last value specified for that index is used.
 Other implementations will also do this for unboxed arrays, but Haskell
-98 requires that for 'Data.Array.Array' the value at such indices is bottom.)
+98 requires that for 'Array' the value at such indices is bottom.)
 
 For most array types, this operation is O(/n/) where /n/ is the size
 of the array.  However, the 'Data.Array.Diff.DiffArray' type provides
@@ -436,16 +444,13 @@ unsafeArrayUArray (l,u) ies default_elem = do
     sequence_ [unsafeWrite marr i e | (i, e) <- ies]
     unsafeFreezeSTUArray marr
 
-#ifdef __GLASGOW_HASKELL__
 {-# INLINE unsafeFreezeSTUArray #-}
 unsafeFreezeSTUArray :: STUArray s i e -> ST s (UArray i e)
+#if __GLASGOW_HASKELL__
 unsafeFreezeSTUArray (STUArray l u n marr#) = ST $ \s1# ->
     case unsafeFreezeByteArray# marr# s1# of { (# s2#, arr# #) ->
     (# s2#, UArray l u n arr# #) }
-#endif
-
-#ifdef __HUGS__
-unsafeFreezeSTUArray :: STUArray s i e -> ST s (UArray i e)
+#elif __HUGS__
 unsafeFreezeSTUArray (STUArray l u n marr) = do
     arr <- unsafeFreezeMutableByteArray marr
     return (UArray l u n arr)
@@ -473,8 +478,8 @@ unsafeAccumUArray f arr ies = do
 {-# INLINE unsafeAccumArrayUArray #-}
 unsafeAccumArrayUArray :: (MArray (STUArray s) e (ST s), Ix i)
                        => (e -> e' -> e) -> e -> (i,i) -> [(Int, e')] -> ST s (UArray i e)
-unsafeAccumArrayUArray f init (l,u) ies = do
-    marr <- newArray (l,u) init
+unsafeAccumArrayUArray f initialValue (l,u) ies = do
+    marr <- newArray (l,u) initialValue
     sequence_ [do
         old <- unsafeRead marr i
         unsafeWrite marr i (f old new)
@@ -553,7 +558,7 @@ instance IArray UArray Bool where
     {-# INLINE unsafeAccum #-}
     unsafeAccum f arr ies = runST (unsafeAccumUArray f arr ies)
     {-# INLINE unsafeAccumArray #-}
-    unsafeAccumArray f init lu ies = runST (unsafeAccumArrayUArray f init lu ies)
+    unsafeAccumArray f initialValue lu ies = runST (unsafeAccumArrayUArray f initialValue lu ies)
 
 instance IArray UArray Char where
     {-# INLINE bounds #-}
@@ -574,7 +579,7 @@ instance IArray UArray Char where
     {-# INLINE unsafeAccum #-}
     unsafeAccum f arr ies = runST (unsafeAccumUArray f arr ies)
     {-# INLINE unsafeAccumArray #-}
-    unsafeAccumArray f init lu ies = runST (unsafeAccumArrayUArray f init lu ies)
+    unsafeAccumArray f initialValue lu ies = runST (unsafeAccumArrayUArray f initialValue lu ies)
 
 instance IArray UArray Int where
     {-# INLINE bounds #-}
@@ -595,7 +600,7 @@ instance IArray UArray Int where
     {-# INLINE unsafeAccum #-}
     unsafeAccum f arr ies = runST (unsafeAccumUArray f arr ies)
     {-# INLINE unsafeAccumArray #-}
-    unsafeAccumArray f init lu ies = runST (unsafeAccumArrayUArray f init lu ies)
+    unsafeAccumArray f initialValue lu ies = runST (unsafeAccumArrayUArray f initialValue lu ies)
 
 instance IArray UArray Word where
     {-# INLINE bounds #-}
@@ -616,7 +621,7 @@ instance IArray UArray Word where
     {-# INLINE unsafeAccum #-}
     unsafeAccum f arr ies = runST (unsafeAccumUArray f arr ies)
     {-# INLINE unsafeAccumArray #-}
-    unsafeAccumArray f init lu ies = runST (unsafeAccumArrayUArray f init lu ies)
+    unsafeAccumArray f initialValue lu ies = runST (unsafeAccumArrayUArray f initialValue lu ies)
 
 instance IArray UArray (Ptr a) where
     {-# INLINE bounds #-}
@@ -637,7 +642,7 @@ instance IArray UArray (Ptr a) where
     {-# INLINE unsafeAccum #-}
     unsafeAccum f arr ies = runST (unsafeAccumUArray f arr ies)
     {-# INLINE unsafeAccumArray #-}
-    unsafeAccumArray f init lu ies = runST (unsafeAccumArrayUArray f init lu ies)
+    unsafeAccumArray f initialValue lu ies = runST (unsafeAccumArrayUArray f initialValue lu ies)
 
 instance IArray UArray (FunPtr a) where
     {-# INLINE bounds #-}
@@ -658,7 +663,7 @@ instance IArray UArray (FunPtr a) where
     {-# INLINE unsafeAccum #-}
     unsafeAccum f arr ies = runST (unsafeAccumUArray f arr ies)
     {-# INLINE unsafeAccumArray #-}
-    unsafeAccumArray f init lu ies = runST (unsafeAccumArrayUArray f init lu ies)
+    unsafeAccumArray f initialValue lu ies = runST (unsafeAccumArrayUArray f initialValue lu ies)
 
 instance IArray UArray Float where
     {-# INLINE bounds #-}
@@ -679,7 +684,7 @@ instance IArray UArray Float where
     {-# INLINE unsafeAccum #-}
     unsafeAccum f arr ies = runST (unsafeAccumUArray f arr ies)
     {-# INLINE unsafeAccumArray #-}
-    unsafeAccumArray f init lu ies = runST (unsafeAccumArrayUArray f init lu ies)
+    unsafeAccumArray f initialValue lu ies = runST (unsafeAccumArrayUArray f initialValue lu ies)
 
 instance IArray UArray Double where
     {-# INLINE bounds #-}
@@ -700,7 +705,7 @@ instance IArray UArray Double where
     {-# INLINE unsafeAccum #-}
     unsafeAccum f arr ies = runST (unsafeAccumUArray f arr ies)
     {-# INLINE unsafeAccumArray #-}
-    unsafeAccumArray f init lu ies = runST (unsafeAccumArrayUArray f init lu ies)
+    unsafeAccumArray f initialValue lu ies = runST (unsafeAccumArrayUArray f initialValue lu ies)
 
 instance IArray UArray (StablePtr a) where
     {-# INLINE bounds #-}
@@ -721,9 +726,10 @@ instance IArray UArray (StablePtr a) where
     {-# INLINE unsafeAccum #-}
     unsafeAccum f arr ies = runST (unsafeAccumUArray f arr ies)
     {-# INLINE unsafeAccumArray #-}
-    unsafeAccumArray f init lu ies = runST (unsafeAccumArrayUArray f init lu ies)
+    unsafeAccumArray f initialValue lu ies = runST (unsafeAccumArrayUArray f initialValue lu ies)
 
 -- bogus StablePtr value for initialising a UArray of StablePtr.
+nullStablePtr :: StablePtr a
 #ifdef __GLASGOW_HASKELL__
 nullStablePtr = StablePtr (unsafeCoerce# 0#)
 #endif
@@ -750,7 +756,7 @@ instance IArray UArray Int8 where
     {-# INLINE unsafeAccum #-}
     unsafeAccum f arr ies = runST (unsafeAccumUArray f arr ies)
     {-# INLINE unsafeAccumArray #-}
-    unsafeAccumArray f init lu ies = runST (unsafeAccumArrayUArray f init lu ies)
+    unsafeAccumArray f initialValue lu ies = runST (unsafeAccumArrayUArray f initialValue lu ies)
 
 instance IArray UArray Int16 where
     {-# INLINE bounds #-}
@@ -771,7 +777,7 @@ instance IArray UArray Int16 where
     {-# INLINE unsafeAccum #-}
     unsafeAccum f arr ies = runST (unsafeAccumUArray f arr ies)
     {-# INLINE unsafeAccumArray #-}
-    unsafeAccumArray f init lu ies = runST (unsafeAccumArrayUArray f init lu ies)
+    unsafeAccumArray f initialValue lu ies = runST (unsafeAccumArrayUArray f initialValue lu ies)
 
 instance IArray UArray Int32 where
     {-# INLINE bounds #-}
@@ -792,7 +798,7 @@ instance IArray UArray Int32 where
     {-# INLINE unsafeAccum #-}
     unsafeAccum f arr ies = runST (unsafeAccumUArray f arr ies)
     {-# INLINE unsafeAccumArray #-}
-    unsafeAccumArray f init lu ies = runST (unsafeAccumArrayUArray f init lu ies)
+    unsafeAccumArray f initialValue lu ies = runST (unsafeAccumArrayUArray f initialValue lu ies)
 
 instance IArray UArray Int64 where
     {-# INLINE bounds #-}
@@ -813,7 +819,7 @@ instance IArray UArray Int64 where
     {-# INLINE unsafeAccum #-}
     unsafeAccum f arr ies = runST (unsafeAccumUArray f arr ies)
     {-# INLINE unsafeAccumArray #-}
-    unsafeAccumArray f init lu ies = runST (unsafeAccumArrayUArray f init lu ies)
+    unsafeAccumArray f initialValue lu ies = runST (unsafeAccumArrayUArray f initialValue lu ies)
 
 instance IArray UArray Word8 where
     {-# INLINE bounds #-}
@@ -834,7 +840,7 @@ instance IArray UArray Word8 where
     {-# INLINE unsafeAccum #-}
     unsafeAccum f arr ies = runST (unsafeAccumUArray f arr ies)
     {-# INLINE unsafeAccumArray #-}
-    unsafeAccumArray f init lu ies = runST (unsafeAccumArrayUArray f init lu ies)
+    unsafeAccumArray f initialValue lu ies = runST (unsafeAccumArrayUArray f initialValue lu ies)
 
 instance IArray UArray Word16 where
     {-# INLINE bounds #-}
@@ -855,7 +861,7 @@ instance IArray UArray Word16 where
     {-# INLINE unsafeAccum #-}
     unsafeAccum f arr ies = runST (unsafeAccumUArray f arr ies)
     {-# INLINE unsafeAccumArray #-}
-    unsafeAccumArray f init lu ies = runST (unsafeAccumArrayUArray f init lu ies)
+    unsafeAccumArray f initialValue lu ies = runST (unsafeAccumArrayUArray f initialValue lu ies)
 
 instance IArray UArray Word32 where
     {-# INLINE bounds #-}
@@ -876,7 +882,7 @@ instance IArray UArray Word32 where
     {-# INLINE unsafeAccum #-}
     unsafeAccum f arr ies = runST (unsafeAccumUArray f arr ies)
     {-# INLINE unsafeAccumArray #-}
-    unsafeAccumArray f init lu ies = runST (unsafeAccumArrayUArray f init lu ies)
+    unsafeAccumArray f initialValue lu ies = runST (unsafeAccumArrayUArray f initialValue lu ies)
 
 instance IArray UArray Word64 where
     {-# INLINE bounds #-}
@@ -897,7 +903,7 @@ instance IArray UArray Word64 where
     {-# INLINE unsafeAccum #-}
     unsafeAccum f arr ies = runST (unsafeAccumUArray f arr ies)
     {-# INLINE unsafeAccumArray #-}
-    unsafeAccumArray f init lu ies = runST (unsafeAccumArrayUArray f init lu ies)
+    unsafeAccumArray f initialValue lu ies = runST (unsafeAccumArrayUArray f initialValue lu ies)
 
 instance (Ix ix, Eq e, IArray UArray e) => Eq (UArray ix e) where
     (==) = eqUArray
@@ -954,10 +960,10 @@ class (Monad m) => MArray a e m where
 	-- The INLINE is crucial, because until we know at least which monad 	
 	-- we are in, the code below allocates like crazy.  So inline it,
 	-- in the hope that the context will know the monad.
-    newArray (l,u) init = do
+    newArray (l,u) initialValue = do
         let n = safeRangeSize (l,u)
         marr <- unsafeNewArray_ (l,u)
-        sequence_ [unsafeWrite marr i init | i <- [0 .. n - 1]]
+        sequence_ [unsafeWrite marr i initialValue | i <- [0 .. n - 1]]
         return marr
 
     {-# INLINE unsafeNewArray_ #-}
@@ -981,6 +987,20 @@ class (Monad m) => MArray a e m where
     -- default initialisation with undefined values if we *do* know the
     -- initial value and it is constant for all elements.
 
+instance MArray IOArray e IO where
+#if defined(__HUGS__)
+    getBounds   = return . boundsIOArray
+    getNumElements = return . getNumElementsIOArray
+#elif defined(__GLASGOW_HASKELL__)
+    {-# INLINE getBounds #-}
+    getBounds (IOArray marr) = stToIO $ getBounds marr
+    {-# INLINE getNumElements #-}
+    getNumElements (IOArray marr) = stToIO $ getNumElements marr
+#endif
+    newArray    = newIOArray
+    unsafeRead  = unsafeReadIOArray
+    unsafeWrite = unsafeWriteIOArray
+
 {-# INLINE newListArray #-}
 -- | Constructs a mutable array from a list of initial elements.
 -- The list gives the elements of the array in ascending order
@@ -1016,7 +1036,7 @@ writeArray marr i e = do
 -- | Return a list of all the elements of a mutable array
 getElems :: (MArray a e m, Ix i) => a i e -> m [e]
 getElems marr = do 
-  (l,u) <- getBounds marr
+  (_l, _u) <- getBounds marr
   n <- getNumElements marr
   sequence [unsafeRead marr i | i <- [0 .. n - 1]]
 
@@ -1102,22 +1122,28 @@ INSTANCE_TYPEABLE3(STArray,sTArrayTc,"STArray")
 --    are supported.
 --
 -- An 'STUArray' will generally be more efficient (in terms of both time
--- and space) than the equivalent boxed version ('Data.Array.ST.STArray')
--- with the same element type.  However, 'STUArray' is strict in its
--- elements - so -- don\'t use 'STUArray' if you require the
--- non-strictness that 'Data.Array.ST.STArray' provides.
+-- and space) than the equivalent boxed version ('STArray') with the same
+-- element type.  However, 'STUArray' is strict in its elements - so
+-- don\'t use 'STUArray' if you require the non-strictness that
+-- 'STArray' provides.
 #ifdef __GLASGOW_HASKELL__
-data STUArray s i a = STUArray !i !i !Int (MutableByteArray# s)
+data STUArray s i e = STUArray !i !i !Int (MutableByteArray# s)
 #endif
 #ifdef __HUGS__
-data STUArray s i a = STUArray !i !i !Int !(MutableByteArray s)
+data STUArray s i e = STUArray !i !i !Int !(MutableByteArray s)
 #endif
 
 INSTANCE_TYPEABLE3(STUArray,stUArrayTc,"STUArray")
 
+#ifdef __GLASGOW_HASKELL__
 instance Eq (STUArray s i e) where
     STUArray _ _ _ arr1# == STUArray _ _ _ arr2# =
         sameMutableByteArray# arr1# arr2#
+#endif
+#ifdef __HUGS__
+instance Eq (STUArray s i e) where
+    STUArray _ _ _ arr1 == STUArray _ _ _ arr2 = arr1 == arr2
+#endif
 
 #ifdef __GLASGOW_HASKELL__
 {-# INLINE unsafeNewArraySTUArray_ #-}
@@ -1137,7 +1163,7 @@ instance MArray (STUArray s) Bool (ST s) where
     {-# INLINE getNumElements #-}
     getNumElements (STUArray _ _ n _) = return n
     {-# INLINE newArray #-}
-    newArray (l,u) init = ST $ \s1# ->
+    newArray (l,u) initialValue = ST $ \s1# ->
         case safeRangeSize (l,u)            of { n@(I# n#) ->
         case newByteArray# (bOOL_SCALE n#) s1# of { (# s2#, marr# #) ->
         case bOOL_WORD_SCALE n#         of { n'# ->
@@ -1148,11 +1174,11 @@ instance MArray (STUArray s) Bool (ST s) where
         case loop 0# s2#                of { s3# ->
         (# s3#, STUArray l u n marr# #) }}}}
       where
-        W# e# = if init then maxBound else 0
+        W# e# = if initialValue then maxBound else 0
     {-# INLINE unsafeNewArray_ #-}
     unsafeNewArray_ (l,u) = unsafeNewArraySTUArray_ (l,u) bOOL_SCALE
     {-# INLINE newArray_ #-}
-    newArray_ bounds = newArray bounds False
+    newArray_ arrBounds = newArray arrBounds False
     {-# INLINE unsafeRead #-}
     unsafeRead (STUArray _ _ _ marr#) (I# i#) = ST $ \s1# ->
         case readWordArray# marr# (bOOL_INDEX i#) s1# of { (# s2#, e# #) ->
@@ -1174,7 +1200,7 @@ instance MArray (STUArray s) Char (ST s) where
     {-# INLINE unsafeNewArray_ #-}
     unsafeNewArray_ (l,u) = unsafeNewArraySTUArray_ (l,u) (*# 4#)
     {-# INLINE newArray_ #-}
-    newArray_ bounds = newArray bounds (chr 0)
+    newArray_ arrBounds = newArray arrBounds (chr 0)
     {-# INLINE unsafeRead #-}
     unsafeRead (STUArray _ _ _ marr#) (I# i#) = ST $ \s1# ->
         case readWideCharArray# marr# i# s1# of { (# s2#, e# #) ->
@@ -1192,7 +1218,7 @@ instance MArray (STUArray s) Int (ST s) where
     {-# INLINE unsafeNewArray_ #-}
     unsafeNewArray_ (l,u) = unsafeNewArraySTUArray_ (l,u) wORD_SCALE
     {-# INLINE newArray_ #-}
-    newArray_ bounds = newArray bounds 0
+    newArray_ arrBounds = newArray arrBounds 0
     {-# INLINE unsafeRead #-}
     unsafeRead (STUArray _ _ _ marr#) (I# i#) = ST $ \s1# ->
         case readIntArray# marr# i# s1# of { (# s2#, e# #) ->
@@ -1210,7 +1236,7 @@ instance MArray (STUArray s) Word (ST s) where
     {-# INLINE unsafeNewArray_ #-}
     unsafeNewArray_ (l,u) = unsafeNewArraySTUArray_ (l,u) wORD_SCALE
     {-# INLINE newArray_ #-}
-    newArray_ bounds = newArray bounds 0
+    newArray_ arrBounds = newArray arrBounds 0
     {-# INLINE unsafeRead #-}
     unsafeRead (STUArray _ _ _ marr#) (I# i#) = ST $ \s1# ->
         case readWordArray# marr# i# s1# of { (# s2#, e# #) ->
@@ -1228,7 +1254,7 @@ instance MArray (STUArray s) (Ptr a) (ST s) where
     {-# INLINE unsafeNewArray_ #-}
     unsafeNewArray_ (l,u) = unsafeNewArraySTUArray_ (l,u) wORD_SCALE
     {-# INLINE newArray_ #-}
-    newArray_ bounds = newArray bounds nullPtr
+    newArray_ arrBounds = newArray arrBounds nullPtr
     {-# INLINE unsafeRead #-}
     unsafeRead (STUArray _ _ _ marr#) (I# i#) = ST $ \s1# ->
         case readAddrArray# marr# i# s1# of { (# s2#, e# #) ->
@@ -1246,7 +1272,7 @@ instance MArray (STUArray s) (FunPtr a) (ST s) where
     {-# INLINE unsafeNewArray_ #-}
     unsafeNewArray_ (l,u) = unsafeNewArraySTUArray_ (l,u) wORD_SCALE
     {-# INLINE newArray_ #-}
-    newArray_ bounds = newArray bounds nullFunPtr
+    newArray_ arrBounds = newArray arrBounds nullFunPtr
     {-# INLINE unsafeRead #-}
     unsafeRead (STUArray _ _ _ marr#) (I# i#) = ST $ \s1# ->
         case readAddrArray# marr# i# s1# of { (# s2#, e# #) ->
@@ -1264,7 +1290,7 @@ instance MArray (STUArray s) Float (ST s) where
     {-# INLINE unsafeNewArray_ #-}
     unsafeNewArray_ (l,u) = unsafeNewArraySTUArray_ (l,u) fLOAT_SCALE
     {-# INLINE newArray_ #-}
-    newArray_ bounds = newArray bounds 0
+    newArray_ arrBounds = newArray arrBounds 0
     {-# INLINE unsafeRead #-}
     unsafeRead (STUArray _ _ _ marr#) (I# i#) = ST $ \s1# ->
         case readFloatArray# marr# i# s1# of { (# s2#, e# #) ->
@@ -1282,7 +1308,7 @@ instance MArray (STUArray s) Double (ST s) where
     {-# INLINE unsafeNewArray_ #-}
     unsafeNewArray_ (l,u) = unsafeNewArraySTUArray_ (l,u) dOUBLE_SCALE
     {-# INLINE newArray_ #-}
-    newArray_ bounds = newArray bounds 0
+    newArray_ arrBounds = newArray arrBounds 0
     {-# INLINE unsafeRead #-}
     unsafeRead (STUArray _ _ _ marr#) (I# i#) = ST $ \s1# ->
         case readDoubleArray# marr# i# s1# of { (# s2#, e# #) ->
@@ -1300,7 +1326,7 @@ instance MArray (STUArray s) (StablePtr a) (ST s) where
     {-# INLINE unsafeNewArray_ #-}
     unsafeNewArray_ (l,u) = unsafeNewArraySTUArray_ (l,u) wORD_SCALE
     {-# INLINE newArray_ #-}
-    newArray_ bounds = newArray bounds (castPtrToStablePtr nullPtr)
+    newArray_ arrBounds = newArray arrBounds (castPtrToStablePtr nullPtr)
     {-# INLINE unsafeRead #-}
     unsafeRead (STUArray _ _ _ marr#) (I# i#) = ST $ \s1# ->
         case readStablePtrArray# marr# i# s1# of { (# s2#, e# #) ->
@@ -1318,7 +1344,7 @@ instance MArray (STUArray s) Int8 (ST s) where
     {-# INLINE unsafeNewArray_ #-}
     unsafeNewArray_ (l,u) = unsafeNewArraySTUArray_ (l,u) (\x -> x)
     {-# INLINE newArray_ #-}
-    newArray_ bounds = newArray bounds 0
+    newArray_ arrBounds = newArray arrBounds 0
     {-# INLINE unsafeRead #-}
     unsafeRead (STUArray _ _ _ marr#) (I# i#) = ST $ \s1# ->
         case readInt8Array# marr# i# s1# of { (# s2#, e# #) ->
@@ -1336,7 +1362,7 @@ instance MArray (STUArray s) Int16 (ST s) where
     {-# INLINE unsafeNewArray_ #-}
     unsafeNewArray_ (l,u) = unsafeNewArraySTUArray_ (l,u) (*# 2#)
     {-# INLINE newArray_ #-}
-    newArray_ bounds = newArray bounds 0
+    newArray_ arrBounds = newArray arrBounds 0
     {-# INLINE unsafeRead #-}
     unsafeRead (STUArray _ _ _ marr#) (I# i#) = ST $ \s1# ->
         case readInt16Array# marr# i# s1# of { (# s2#, e# #) ->
@@ -1354,7 +1380,7 @@ instance MArray (STUArray s) Int32 (ST s) where
     {-# INLINE unsafeNewArray_ #-}
     unsafeNewArray_ (l,u) = unsafeNewArraySTUArray_ (l,u) (*# 4#)
     {-# INLINE newArray_ #-}
-    newArray_ bounds = newArray bounds 0
+    newArray_ arrBounds = newArray arrBounds 0
     {-# INLINE unsafeRead #-}
     unsafeRead (STUArray _ _ _ marr#) (I# i#) = ST $ \s1# ->
         case readInt32Array# marr# i# s1# of { (# s2#, e# #) ->
@@ -1372,7 +1398,7 @@ instance MArray (STUArray s) Int64 (ST s) where
     {-# INLINE unsafeNewArray_ #-}
     unsafeNewArray_ (l,u) = unsafeNewArraySTUArray_ (l,u) (*# 8#)
     {-# INLINE newArray_ #-}
-    newArray_ bounds = newArray bounds 0
+    newArray_ arrBounds = newArray arrBounds 0
     {-# INLINE unsafeRead #-}
     unsafeRead (STUArray _ _ _ marr#) (I# i#) = ST $ \s1# -> 
         case readInt64Array# marr# i# s1# of { (# s2#, e# #) ->
@@ -1390,7 +1416,7 @@ instance MArray (STUArray s) Word8 (ST s) where
     {-# INLINE unsafeNewArray_ #-}
     unsafeNewArray_ (l,u) = unsafeNewArraySTUArray_ (l,u) (\x -> x)
     {-# INLINE newArray_ #-}
-    newArray_ bounds = newArray bounds 0
+    newArray_ arrBounds = newArray arrBounds 0
     {-# INLINE unsafeRead #-}
     unsafeRead (STUArray _ _ _ marr#) (I# i#) = ST $ \s1# ->
         case readWord8Array# marr# i# s1# of { (# s2#, e# #) ->
@@ -1408,7 +1434,7 @@ instance MArray (STUArray s) Word16 (ST s) where
     {-# INLINE unsafeNewArray_ #-}
     unsafeNewArray_ (l,u) = unsafeNewArraySTUArray_ (l,u) (*# 2#)
     {-# INLINE newArray_ #-}
-    newArray_ bounds = newArray bounds 0
+    newArray_ arrBounds = newArray arrBounds 0
     {-# INLINE unsafeRead #-}
     unsafeRead (STUArray _ _ _ marr#) (I# i#) = ST $ \s1# ->
         case readWord16Array# marr# i# s1# of { (# s2#, e# #) ->
@@ -1426,7 +1452,7 @@ instance MArray (STUArray s) Word32 (ST s) where
     {-# INLINE unsafeNewArray_ #-}
     unsafeNewArray_ (l,u) = unsafeNewArraySTUArray_ (l,u) (*# 4#)
     {-# INLINE newArray_ #-}
-    newArray_ bounds = newArray bounds 0
+    newArray_ arrBounds = newArray arrBounds 0
     {-# INLINE unsafeRead #-}
     unsafeRead (STUArray _ _ _ marr#) (I# i#) = ST $ \s1# ->
         case readWord32Array# marr# i# s1# of { (# s2#, e# #) ->
@@ -1444,7 +1470,7 @@ instance MArray (STUArray s) Word64 (ST s) where
     {-# INLINE unsafeNewArray_ #-}
     unsafeNewArray_ (l,u) = unsafeNewArraySTUArray_ (l,u) (*# 8#)
     {-# INLINE newArray_ #-}
-    newArray_ bounds = newArray bounds 0
+    newArray_ arrBounds = newArray arrBounds 0
     {-# INLINE unsafeRead #-}
     unsafeRead (STUArray _ _ _ marr#) (I# i#) = ST $ \s1# ->
         case readWord64Array# marr# i# s1# of { (# s2#, e# #) ->
@@ -1674,12 +1700,10 @@ freeze :: (Ix i, MArray a e m, IArray b e) => a i e -> m (b i e)
 freeze marr = do
   (l,u) <- getBounds marr
   n <- getNumElements marr
-  ies <- sequence [do e <- unsafeRead marr i; return (i,e)
-                   | i <- [0 .. n - 1]]
-  -- The old array may be lying about the number of elements in
-  -- (l,u), so recalculate it to be safe.
-  let n' = safeRangeSize (l,u)
-  return (unsafeArray (l,u) ies)
+  es <- mapM (unsafeRead marr) [0 .. n - 1]
+  -- The old array and index might not be well-behaved, so we need to
+  -- use the safe array creation function here.
+  return (listArray (l,u) es)
 
 #ifdef __GLASGOW_HASKELL__
 freezeSTUArray :: Ix i => STUArray s i e -> ST s (UArray i e)
@@ -1757,8 +1781,8 @@ thaw arr = case bounds arr of
               | i <- [0 .. n - 1]]
     return marr
 
-#ifdef __GLASGOW_HASKELL__
 thawSTUArray :: Ix i => UArray i e -> ST s (STUArray s i e)
+#if __GLASGOW_HASKELL__
 thawSTUArray (UArray l u n arr#) = ST $ \s1# ->
     case sizeofByteArray# arr#          of { n# ->
     case newByteArray# n# s1#           of { (# s2#, marr# #) ->
@@ -1774,10 +1798,7 @@ foreign import ccall unsafe "memcpy"
 "thaw/STArray"  thaw = ArrST.thawSTArray
 "thaw/STUArray" thaw = thawSTUArray
     #-}
-#endif /* __GLASGOW_HASKELL__ */
-
-#ifdef __HUGS__
-thawSTUArray :: Ix i => UArray i e -> ST s (STUArray s i e)
+#elif __HUGS__
 thawSTUArray (UArray l u n arr) = do
     marr <- thawByteArray arr
     return (STUArray l u n marr)
@@ -1835,18 +1856,49 @@ unsafeThawSTUArray (UArray l u n marr#) =
 "unsafeThaw/STArray"    unsafeThaw = ArrST.unsafeThawSTArray
 "unsafeThaw/STUArray"   unsafeThaw = unsafeThawSTUArray
     #-}
+
+{-# INLINE unsafeThawIOArray #-}
+unsafeThawIOArray :: Ix ix => Arr.Array ix e -> IO (IOArray ix e)
+unsafeThawIOArray arr = stToIO $ do
+    marr <- ArrST.unsafeThawSTArray arr
+    return (IOArray marr)
+
+{-# RULES
+"unsafeThaw/IOArray"  unsafeThaw = unsafeThawIOArray
+    #-}
+
+thawIOArray :: Ix ix => Arr.Array ix e -> IO (IOArray ix e)
+thawIOArray arr = stToIO $ do
+    marr <- ArrST.thawSTArray arr
+    return (IOArray marr)
+
+{-# RULES
+"thaw/IOArray"  thaw = thawIOArray
+    #-}
+
+freezeIOArray :: Ix ix => IOArray ix e -> IO (Arr.Array ix e)
+freezeIOArray (IOArray marr) = stToIO (ArrST.freezeSTArray marr)
+
+{-# RULES
+"freeze/IOArray"  freeze = freezeIOArray
+    #-}
+
+{-# INLINE unsafeFreezeIOArray #-}
+unsafeFreezeIOArray :: Ix ix => IOArray ix e -> IO (Arr.Array ix e)
+unsafeFreezeIOArray (IOArray marr) = stToIO (ArrST.unsafeFreezeSTArray marr)
+
+{-# RULES
+"unsafeFreeze/IOArray"  unsafeFreeze = unsafeFreezeIOArray
+    #-}
 #endif /* __GLASGOW_HASKELL__ */
 
 -- | Casts an 'STUArray' with one element type into one with a
 -- different element type.  All the elements of the resulting array
 -- are undefined (unless you know what you\'re doing...).
 
-#ifdef __GLASGOW_HASKELL__
 castSTUArray :: STUArray s ix a -> ST s (STUArray s ix b)
+#if __GLASGOW_HASKELL__
 castSTUArray (STUArray l u n marr#) = return (STUArray l u n marr#)
-#endif
-
-#ifdef __HUGS__
-castSTUArray :: STUArray s ix a -> ST s (STUArray s ix b)
+#elif __HUGS__
 castSTUArray (STUArray l u n marr) = return (STUArray l u n marr)
 #endif
