@@ -30,7 +30,7 @@
 #include "ThreadLabels.h"
 #include "BlockAlloc.h"
 #include "Trace.h"
-#include "RtsTypeable.h"
+#include "RtsGlobals.h"
 #include "Stable.h"
 #include "Hpc.h"
 #include "FileLock.h"
@@ -71,6 +71,9 @@
 #ifdef HAVE_SIGNAL_H
 #include <signal.h>
 #endif
+#ifdef HAVE_UNISTD_H
+#include <unistd.h>
+#endif
 
 #if USE_PAPI
 #include "Papi.h"
@@ -140,17 +143,20 @@ hs_init(int *argc, char **argv[])
 	return;
     }
 
-#if defined(DEBUG)
-    /* Start off by initialising the allocator debugging so we can
-     * use it anywhere */
-    initAllocator();
-#endif
+    /* Initialise the stats department, phase 0 */
+    initStats0();
 
     /* Next we do is grab the start time...just in case we're
      * collecting timing statistics.
      */
     stat_startInit();
 
+#if defined(DEBUG)
+    /* Start off by initialising the allocator debugging so we can
+     * use it anywhere */
+    initAllocator();
+#endif
+
 #ifdef PAR
     /*
      * The parallel system needs to be initialised and synchronised before
@@ -167,30 +173,6 @@ hs_init(int *argc, char **argv[])
     argv++; argc--;
 #endif
 
-    /* Initialise the performance tracking library */
-#ifdef USE_PAPI
-    {
-	int ver;
-	if ((ver = PAPI_library_init(PAPI_VER_CURRENT)) != PAPI_VER_CURRENT) {
-	    if (ver > 0) {
-		errorBelch("PAPI_library_init: wrong version: %x", ver);
-		stg_exit(EXIT_FAILURE);
-	    } else {
-		sysErrorBelch("PAPI_library_init");
-		stg_exit(EXIT_FAILURE);
-	    }
-	}
-    }
-#ifdef THREADED_RTS
-    {
-	int err;
-	if ((err = PAPI_thread_init(osThreadId)) < 0) {
-	    barf("PAPI_thread_init: %d",err);
-	}
-    }
-#endif
-#endif
-
     /* Set the RTS flags to default values. */
 
     initRtsFlagsDefaults();
@@ -205,6 +187,13 @@ hs_init(int *argc, char **argv[])
 	setProgArgv(*argc,*argv);
     }
 
+    /* Initialise the stats department, phase 1 */
+    initStats1();
+
+#ifdef USE_PAPI
+    papi_init();
+#endif
+
     /* initTracing must be after setupRtsFlags() */
     initTracing();
 
@@ -236,8 +225,22 @@ hs_init(int *argc, char **argv[])
     /* initialise the stable pointer table */
     initStablePtrTable();
 
+    /* Add some GC roots for things in the base package that the RTS
+     * knows about.  We don't know whether these turn out to be CAFs
+     * or refer to CAFs, but we have to assume that they might.
+     */
+    getStablePtr((StgPtr)base_GHCziTopHandler_runIO_closure);
+    getStablePtr((StgPtr)base_GHCziTopHandler_runNonIO_closure);
+    getStablePtr((StgPtr)stackOverflow_closure);
+    getStablePtr((StgPtr)heapOverflow_closure);
+    getStablePtr((StgPtr)runFinalizerBatch_closure);
+    getStablePtr((StgPtr)unpackCString_closure);
+    getStablePtr((StgPtr)blockedOnDeadMVar_closure);
+    getStablePtr((StgPtr)nonTermination_closure);
+    getStablePtr((StgPtr)blockedIndefinitely_closure);
+
     /* initialise the shared Typeable store */
-    initTypeableStore();
+    initGlobalStore();
 
     /* initialise file locking, if necessary */
 #if !defined(mingw32_HOST_OS)    
@@ -255,9 +258,6 @@ hs_init(int *argc, char **argv[])
     initTimer();
     startTimer();
 
-    /* Initialise the stats department */
-    initStats();
-
 #if defined(RTS_USER_SIGNALS)
     if (RtsFlags.MiscFlags.install_signal_handlers) {
         /* Initialise the user signal handler set */
@@ -400,11 +400,7 @@ hs_exit_(rtsBool wait_foreign)
     /* start timing the shutdown */
     stat_startExit();
     
-#if defined(RTS_USER_SIGNALS)
-    if (RtsFlags.MiscFlags.install_signal_handlers) {
-        freeSignalHandlers();
-    }
-#endif
+    OnExitHook();
 
 #if defined(THREADED_RTS)
     ioManagerDie();
@@ -412,7 +408,16 @@ hs_exit_(rtsBool wait_foreign)
 
     /* stop all running tasks */
     exitScheduler(wait_foreign);
-    
+
+    /* run C finalizers for all active weak pointers */
+    runAllCFinalizers(weak_ptr_list);
+
+#if defined(RTS_USER_SIGNALS)
+    if (RtsFlags.MiscFlags.install_signal_handlers) {
+        freeSignalHandlers();
+    }
+#endif
+
 #if defined(GRAN)
     /* end_gr_simulation prints global stats if requested -- HWL */
     if (!RtsFlags.GranFlags.GranSimStats.Suppressed)
@@ -457,6 +462,9 @@ hs_exit_(rtsBool wait_foreign)
     PAR_TICKY_PAR_END();
 #endif
 
+    // uninstall signal handlers
+    resetDefaultHandlers();
+
     /* stop timing the shutdown, we're about to print stats */
     stat_endExit();
     
@@ -470,8 +478,7 @@ hs_exit_(rtsBool wait_foreign)
     /* free the tasks */
     freeScheduler();
 
-    /* free shared Typeable store */
-    exitTypeableStore();
+    exitGlobalStore();
 
     /* free file locking tables, if necessary */
 #if !defined(mingw32_HOST_OS)    
@@ -545,10 +552,10 @@ shutdownHaskell(void)
 void
 shutdownHaskellAndExit(int n)
 {
-    if (hs_init_count == 1) {
-	OnExitHook();
-	hs_exit_(rtsFalse);
-        // we're about to exit(), no need to wait for foreign calls to return.
+    // we're about to exit(), no need to wait for foreign calls to return.
+    hs_exit_(rtsFalse);
+
+    if (hs_init_count == 0) {
 #if defined(PAR)
 	/* really exit (stg_exit() would call shutdownParallelSystem() again) */
 	exit(n);
@@ -558,6 +565,15 @@ shutdownHaskellAndExit(int n)
     }
 }
 
+#ifndef mingw32_HOST_OS
+void
+shutdownHaskellAndSignal(int sig)
+{
+    hs_exit_(rtsFalse);
+    kill(getpid(),sig);
+}
+#endif
+
 /* 
  * called from STG-land to exit the program
  */
